// Flocking.h
// Implements basic force functions as discussed by Reynolds' in 1987
// (see e.g. http://www.red3d.com/cwr/boids/)
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef SWARMFLOCKING_SIMULATION_FLOCKING_H
#define SWARMFLOCKING_SIMULATION_FLOCKING_H

#include "../common/Boid.h" // for Boid
#include "../common/Vector.h" // for Vector3f and its methods
#include "../common/Configuration.h" // global settings like world size

#include <immintrin.h>

// Implements Reynold's steering formula:
// steering = desired (towards steer with maximum speed) - current (velocity)
static inline Vector3f reynoldsSteering(const Vector3f &current, Vector3f steer,
                                        std::size_t influenceCount, float weight)
{
    if (influenceCount > 0) {
        steer /= static_cast<float>(influenceCount);
    }

    if (steer.lengthsq() > 0.0f) {
        steer = steer.setLength(configuration.boid.limit.speed) - current;
        steer = steer.limitLength(configuration.boid.limit.force);
    }

    return steer * weight;
}

// correctly determine the difference vector between two boids,
// keeping in mind that the box containing all boids wraps around
inline Vector3f calculateDifference(const Vector3f& lhs, const Vector3f& rhs)
{
    Vector3f result;

    // load vector representations of input data
    __m128 lhs_ = _mm_load_ps((const float *)&lhs),
           rhs_ = _mm_load_ps((const float *)&rhs);

    // mask to extract sign bits from ieee754 float representations:
    // vector with bit representation (1 << 31) => -0.0f for each element
    const __m128 signM  = _mm_set1_ps(-0.0f);

    // the greatest distance in a dimension with the specified extent
    const __m128 limitM = _mm_set1_ps(configuration.worldSize / 2.0f);

    // vector math:
    //       rhs = lhs + b
    // <=>   rhs - lhs = b
    const __m128 directly = _mm_sub_ps(rhs_, lhs_);

    // vector abs: set sign bit to 0 for each element in vector
    // so positive values remain positive, whereas negative values
    // are now positive
    const __m128 absDirectly = _mm_andnot_ps(signM, directly);

    // mask with bit pattern ~0 for elements of absDirectly if absDirectly[i] > limitM[i],
    // otherwise bit pattern 0
    const __m128 indirectlyM = _mm_cmpgt_ps(absDirectly, limitM);

    // 1.0f iff corresponding element in directly is negative, otherwise -1.0f!
    const __m128 factor = _mm_xor_ps(_mm_set1_ps(-1.0f), _mm_and_ps(directly, signM));

    // correct result, iff abs(directly) > (limit / 2.0f)
    const __m128 indirectly = _mm_mul_ps(_mm_sub_ps(_mm_set1_ps(configuration.worldSize),
                                              absDirectly), // limit - std::abs(directly)
                                        factor); // * ((directly < 0.0f) ? 1.0f : -1.0f)

    // store the result in result vector
    _mm_store_ps((float *)&result,
                 _mm_or_ps(_mm_and_ps(   indirectlyM, indirectly), // choose indirectly according to mask
                           _mm_andnot_ps(indirectlyM, directly))); // choose directly according to inverted mask

    return result;
}

// "steer to avoid crowding local flockmates" - Reynolds' 1987
// see Swarm::calculateSteer for usage!
inline Vector3f separation(const Boid &boid, const Boid &other)
{
    // Vector3f difference = boid.position - other.position;
    Vector3f difference = calculateDifference(other.position, boid.position);
    float distance      = difference.length();

    // optimization for difference.normalize() / distance;
    return difference / (distance * distance);
}

// "steer towards the average heading of local flockmates" - Reynolds' 1987
// the averaging part of the formula is accomplished in Swarm::calculateSteer!
inline Vector3f alignment(const Boid &boid, const Boid &other)
{
    return other.velocity;
}

// "steer to move toward the average position of local flockmates" - Reynolds' 1987
inline Vector3f cohesion(const Boid &boid, const Boid &other)
{
    // 'shadow' position - where the other boid is at logically, ignoring the boundaries
    // of the surrounding world
    return boid.position + calculateDifference(boid.position, other.position);
    // return other.position;
}

inline Vector3f finalizeSeparation(const Boid &boid, Vector3f steer,
                                   std::size_t influenceCount, float separationWeight)
{
    return reynoldsSteering(boid.velocity, steer, influenceCount, separationWeight);
}

inline Vector3f finalizeAlignment(const Boid &boid, Vector3f steer,
                                  std::size_t influenceCount, float alignmentWeight)
{
    return reynoldsSteering(boid.velocity, steer, influenceCount, alignmentWeight);
}

inline Vector3f finalizeCohesion(const Boid &boid, Vector3f steer,
                                 std::size_t influenceCount, float cohesionWeight)
{
    // workaround to use reynoldsSteering (see second parameter below!)
    if (influenceCount > 0) {
        steer /= static_cast<float>(influenceCount);
    }

    return reynoldsSteering(boid.velocity, steer - boid.position, 0, cohesionWeight);
}

#endif /* SWARMFLOCKING_SIMULATION_FLOCKING_H */
