// PartialSwarm.cpp
// A part of the total swarm that is aware of its relevant neighbors
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "PartialSwarm.h"

#include <utility> // move, begin, end
#include <cassert> // assert
#include <cmath> // ceil
#include <algorithm> // copy_if
#include <iterator> // back_inserter, distance

#include "../common/Configuration.h"

PartialSwarm::PartialSwarm(Swarm &swarm, std::size_t rank) :
    swarm(swarm)
{
    // get the maximum boid range
    auto maxRange = std::max(configuration.boid.range.separation,
                             std::max(configuration.boid.range.cohesion,
                                      configuration.boid.range.alignment));

    // neighbors of interest in each direction
    if (swarm.maxRank < 4) {
        // TODO: fix formula below
        // Formula below does not work for less than 4 processes
        assert(false);
    } else {
        auto neighborCount = static_cast<unsigned int>(std::ceil((swarm.maxRank * maxRange) / configuration.worldSize));
        neighbors.begin = &swarm.parts[(rank - neighborCount + swarm.maxRank) % swarm.maxRank];
        neighbors.end   = &swarm.parts[(rank + neighborCount + 1) % swarm.maxRank];
    }

    minX = static_cast<float>(rank       * (configuration.worldSize / swarm.maxRank));
    maxX = static_cast<float>((rank + 1) * (configuration.worldSize / swarm.maxRank));

    // determine the boids starting in this part of the swarmm
    for (auto &boid : swarm.boids) {
        if (boid.position.x >= minX && boid.position.x < maxX) {
            boids.emplace_back(boid);
        }
    }
}

// is *this a left neighbor of partialSwarm?
bool PartialSwarm::isLeftNeighborOf(PartialSwarm &partialSwarm) const
{
    for (const auto &a : partialSwarm) {
        if (&a == this) {
            return true;
        } else if (&a == &partialSwarm) {
            return false;
        }
    }
}

bool PartialSwarm::isRightNeighborOf(PartialSwarm &partialSwarm) const
{
    return !isLeftNeighborOf(partialSwarm);
}

// calculates rank of the owner of the partial swarm
int PartialSwarm::tag() const
{
    return this - swarm.parts;
}

PartialSwarm::iterator PartialSwarm::begin()
{
    return { swarm, neighbors.begin };
}

PartialSwarm::iterator PartialSwarm::current()
{
    return { swarm, this };
}

PartialSwarm::iterator PartialSwarm::end()
{
    return { swarm, neighbors.end };
}

PartialSwarm::iterator::iterator(Swarm &swarm, PartialSwarm *current)
    : swarm(swarm), current{current}
{
}

PartialSwarm::iterator &PartialSwarm::iterator::operator++()
{
    if (++current == swarm.parts + swarm.maxRank) {
        current = swarm.parts;
    }

    return *this;
}

PartialSwarm::iterator PartialSwarm::iterator::operator++(int)
{
    PartialSwarm::iterator value{*this};
    operator++();
    return value;
}

PartialSwarm::iterator &PartialSwarm::iterator::operator--()
{
    if (current-- == swarm.parts) {
        current = swarm.parts + swarm.maxRank - 1;
    }

    return *this;
}

PartialSwarm::iterator PartialSwarm::iterator::operator--(int)
{
    PartialSwarm::iterator value{*this};
    operator--();
    return value;
}

PartialSwarm &PartialSwarm::iterator::operator*()
{
    return *current;
}

PartialSwarm *PartialSwarm::iterator::operator->()
{
    return current;
}

bool PartialSwarm::iterator::operator==(const PartialSwarm::iterator &rhs)
{
    return current == rhs.current;
}

bool PartialSwarm::iterator::operator!=(const PartialSwarm::iterator &rhs)
{
    return current != rhs.current;
}
