// PartialSwarm.h
// A part of the total swarm that is aware of its relevant neighbors
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef SWARMFLOCKING_SIMULATION_PARTIALSWARM_H
#define SWARMFLOCKING_SIMULATION_PARTIALSWARM_H

#include <cstdint> // size_t
#include <vector> // vector
#include <cstddef> // ptrdiff_t

#include "Swarm.h"
#include "../common/Boid.h"

// A partial swarm which is aware of his relevant neighbrs
class PartialSwarm
{
public:
    typedef std::ptrdiff_t difference_type;
    typedef std::size_t size_type;
    typedef PartialSwarm value_type;
    typedef PartialSwarm *pointer;
    typedef PartialSwarm &reference;

    // Iterator support (cyclic)
    class iterator :
        public std::iterator<
            std::bidirectional_iterator_tag,
            value_type,
            size_type,
            pointer,
            reference>
    {
    public:
        iterator(Swarm &swarm, PartialSwarm *current);

        iterator &operator++();
        iterator operator++(int);

        iterator &operator--();
        iterator operator--(int);

        PartialSwarm &operator*();
        PartialSwarm *operator->();

        bool operator==(const iterator &rhs);
        bool operator!=(const iterator &rhs);

    private:
        Swarm &swarm;
        PartialSwarm *current;
    };

    // Construct a new partial swarm from its rank in the parent swarm
    PartialSwarm(Swarm &swarm, std::size_t rank);

    // Boids currently in this partial swarm
    std::vector<Boid> boids;

    // calculates rank of the owner of the partial swarm
    int tag() const;
    bool isLeftNeighborOf(PartialSwarm &partialSwarm) const;
    bool isRightNeighborOf(PartialSwarm &partialSwarm) const;

    // minimum x value such that the corresponding boid
    // is in this bucket
    float minX;

    // maximum x value such that the corresponding boid
    // is not in the bucket
    // => Boid in bucket <=> Boid.x >= minX && Boid.x < maxX
    float maxX;

    iterator begin();
    iterator current();
    iterator end();

private:
    // Reference to the full swarm
    Swarm &swarm;

    // Partial swarms in the relevant neighbor range
    struct {
        PartialSwarm *begin;
        PartialSwarm *end;
    } neighbors;
};

#endif /* SWARMFLOCKING_SIMULATION_PARTIALSWARM_H */
