// Simulation.cpp
// Manages the simulation. Responsible for timing, serialisation and
// synchronisation in the MPI variant.
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Simulation.h"
#include "Timer.h"
#include "Swarm.h"
#include "PartialSwarm.h"
#include "PartialForce.h"

#include "../common/Configuration.h"
#include "../common/Vector.h"

#include <cstddef>
#include <cassert>
#include <future>
#include <numeric>
#include <algorithm>
#include <memory> // unique_ptr

#include <mpi.h>

// create a new simulation with a given amount of boids
Simulation::Simulation(std::size_t boidCount, std::size_t predatorCount, MPI_File file,
                       std::size_t steps, int maxRank, int rank) :
    maxRank{maxRank},
    rank{rank},
    swarm{boidCount, predatorCount, maxRank},
    steps{steps},
    file{file}
{
    MPI_Datatypes_init();
}

Simulation::~Simulation()
{
}

static MPI_Aint type_extent(MPI_Datatype type)
{
    MPI_Aint lowerbound, extent;
    MPI_Type_get_extent(type, &lowerbound, &extent);
    return extent;
}

void Simulation::MPI_VECTOR3F_init()
{
    MPI_Datatype types[] = { MPI_FLOAT, MPI_FLOAT, MPI_FLOAT };
    int blocklengths[] = { 1, 1, 1 };
    MPI_Aint displacements[] = { offsetof(Vector3f, x),
                                 offsetof(Vector3f, y),
                                 offsetof(Vector3f, z) };

    // temporary data type, missing the padding data
    MPI_Datatype _MPI_VECTOR3F;
    MPI_Type_create_struct(3, blocklengths, displacements, types, &_MPI_VECTOR3F);

    // resize the data type to account for .padding field (this does not get send!)
    MPI_Type_create_resized(_MPI_VECTOR3F, 0, sizeof(Vector3f), &MPI_VECTOR3F);
    MPI_Type_commit(&MPI_VECTOR3F);

    // Error checking and datatype validation
    // The actual extent should be as large as the underlying data...
    assert(type_extent(MPI_VECTOR3F) == sizeof(Vector3f));

    // but the padding field does not need to be sent
    int MPI_VECTOR3F_size;
    MPI_Type_size(MPI_VECTOR3F, &MPI_VECTOR3F_size);
    assert(MPI_VECTOR3F_size == (sizeof(float) * 3));
}

void Simulation::MPI_PARTIALFORCE_init()
{
    MPI_Datatype types[] = { MPI_VECTOR3F, MPI_UNSIGNED };
    int blocklengths[] = { 1, 1 };
    MPI_Aint displacements[] = { offsetof(PartialForce, vec),
                                 offsetof(PartialForce, influenceCount) };

    MPI_Type_create_struct(2, blocklengths, displacements, types, &MPI_PARTIALFORCE);
    MPI_Type_commit(&MPI_PARTIALFORCE);

    // Error checkingxs
    assert(type_extent(MPI_PARTIALFORCE) == sizeof(PartialForce));
}

void Simulation::MPI_BOID_init()
{
    MPI_Datatype types[] = { MPI_VECTOR3F, MPI_VECTOR3F };
    int blocklengths[] = { 1, 1 };
    MPI_Aint displacements[] = { offsetof(Boid, position),
                                 offsetof(Boid, velocity) };
    MPI_Datatype _MPI_BOID;
    MPI_Type_create_struct(2, blocklengths, displacements, types, &_MPI_BOID);

    MPI_Type_create_resized(_MPI_BOID, 0, sizeof(Boid), &MPI_BOID);
    MPI_Type_commit(&MPI_BOID);

    // Error checking and datatype validation
    assert(type_extent(MPI_BOID) == sizeof(Boid));

    int MPI_VECTOR3F_size, MPI_BOID_size;
    MPI_Type_size(MPI_VECTOR3F, &MPI_VECTOR3F_size);
    MPI_Type_size(MPI_BOID, &MPI_BOID_size);

    assert(MPI_BOID_size == (2 * MPI_VECTOR3F_size));
}

void Simulation::MPI_BOID_THIN_init()
{
    MPI_Datatype types[] = { MPI_VECTOR3F }; // just the position
    int blocklengths[] = { 1 };
    MPI_Aint displacements[] = { offsetof(Boid, position) };

    // temporary data type, missing velocity field!
    MPI_Datatype _MPI_BOID_THIN;
    MPI_Type_create_struct(1, blocklengths, displacements, types, &_MPI_BOID_THIN);

    MPI_Type_create_resized(_MPI_BOID_THIN, 0, sizeof(Boid), &MPI_BOID_THIN);
    MPI_Type_commit(&MPI_BOID_THIN);

    // Error checking and datatype validation
    assert(type_extent(MPI_BOID_THIN) == sizeof(Boid));

    int MPI_VECTOR3F_size, MPI_BOID_THIN_size;
    MPI_Type_size(MPI_VECTOR3F, &MPI_VECTOR3F_size);
    MPI_Type_size(MPI_BOID_THIN, &MPI_BOID_THIN_size);

    assert(MPI_VECTOR3F_size == MPI_BOID_THIN_size);
}

void Simulation::MPI_Datatypes_init()
{
    MPI_VECTOR3F_init();
    MPI_PARTIALFORCE_init();
    MPI_BOID_init();
    MPI_BOID_THIN_init();
}

// main loop
void Simulation::run()
{
    // Run the simualation for a given amount of steps
    for (std::size_t index = 0; index < steps; ++index) {
        step();
    }
}

// runs a single step of swarm behaviour
// returns number of nanoseconds describing how long the computations (excluding comms) took
void Simulation::step()
{
    // Update the local partial swarm and its neighbors
    swarm.updatePartialSwarms(rank);

    // Calculate the next swarm update
    #pragma omp parallel for
    for (std::size_t i = 0; i < swarm.parts[rank].boids.size(); ++i) {
        Boid &boid = swarm.parts[rank].boids[i];

        if (!swarm.avoidPredators(boid)) {
            swarm.calculateSteer(boid, rank);
        }
    }

    // calculate partial forces, to be
    // redistributed to all other processes during synchronisation
    for (auto &predator : swarm.predators) {
        swarm.calculatePartialPredatorForce(predator, rank);
    }

    // Update the positions
    swarm.updatePositions(swarm.parts[rank].boids, configuration.boid.limit.speed);

    // predator positions get updated during synchronize, as necessary data from other processes
    // is still missing at this point!

    synchronize();
    serialize();
}

bool Simulation::isRoot()
{
    return rank == 0;
}

void Simulation::synchronize()
{
    PartialSwarm &currentPart = swarm.parts[rank];

    // Deal with predator data! The partial forces need to be distributed to every other process,
    // then these partial forces need to be summed up
    // We are using Allgatherv rather than Reduce, because reduce requires an associative operation
    // and floating point arithmetic is not associative (sigh...)
    std::vector<PartialForce> pforces(maxRank * swarm.predatorForces.size());
    MPI_Request predatorReq;

    MPI_Iallgather(swarm.predatorForces.data(), swarm.predatorForces.size(), MPI_PARTIALFORCE,
                   pforces.data(), swarm.predatorForces.size(), MPI_PARTIALFORCE, MPI_COMM_WORLD,
                   &predatorReq);

    // temps contains pointers to temporary vectors that contain
    // boids to send to neighbors. Since we are using asynchronous sends,
    // we can't just declare them locally.
    std::vector< std::unique_ptr<std::vector<Boid> >> temps;


    // Send local boid data to all neighbors
    for (auto &neighbor : currentPart) {
        // We don't need to send data to ourself
        if (neighbor.tag() != rank) {
            static const float maxDistance = std::max(std::max(configuration.boid.range.cohesion,
                                                               configuration.boid.range.separation),
                                                      configuration.boid.range.alignment);

            // Data to be sent to neighbor
            temps.emplace_back(new std::vector<Boid>());
            auto &temp = temps.back();

            // Filter boids so only necessary ones are sent!
            // if 'neighbor' is one of the neighbors to the left of the current part
            if (neighbor.isLeftNeighborOf(currentPart)) {
                const float maxX = neighbor.maxX + maxDistance
                                 - ((neighbor.tag() > rank) ? configuration.worldSize : 0.0f);

                assert(maxX >= currentPart.minX);

                for (const Boid &boid : currentPart.boids) {
                    if (   (    (boid.position.x >= currentPart.minX)
                            &&  (boid.position.x < maxX))
                        || (    (boid.position.x >= neighbor.minX)
                            &&  (boid.position.x < neighbor.maxX))) {
                        temp->emplace_back(boid);
                    }
                }
            } else /* if (neighbor.isRightNeighborOf(currentPart)) */ {
                const float minX = (neighbor.minX - maxDistance)
                                 + ((neighbor.tag() < rank) ? configuration.worldSize : 0.0f);

                assert(minX <= currentPart.maxX);

                for (const Boid &boid : currentPart.boids) {
                    if (    (    (boid.position.x >= minX)
                             &&  (boid.position.x < currentPart.maxX))
                        ||  (    (boid.position.x >= neighbor.minX)
                             &&  (boid.position.x < neighbor.maxX))) {
                        temp->emplace_back(boid);
                    }
                }
            }

            MPI_Request req;
            MPI_Isend(temp->data(), temp->size(), MPI_BOID, neighbor.tag(), currentPart.tag(),
                      MPI_COMM_WORLD, &req);
            MPI_Request_free(&req);
        }
    }

    // Receive updated boid data from neighbors
    for (auto &neighbor : currentPart) {
        if (neighbor.tag() != rank) {
            MPI_Status status;

            MPI_Probe(neighbor.tag(), neighbor.tag(), MPI_COMM_WORLD, &status);

            int count;
            MPI_Get_count(&status, MPI_BOID, &count);

            neighbor.boids.resize(count);
            MPI_Recv(neighbor.boids.data(), count, MPI_BOID, neighbor.tag(), neighbor.tag(),
                     MPI_COMM_WORLD, &status);
        }
    }

    // Continue predator synchronisation:
    // Wait for predator synchronisation to end
    MPI_Wait(&predatorReq, MPI_STATUS_IGNORE);

    // Sum up all the different PartialForces we received. Keep in mind that they are now stored in strides,
    // meaning the first j elements of pforces are all the partial forces coming from process 0, the next j
    // elements come from process 1, ...
    for (std::size_t outer = 0; outer < swarm.predatorForces.size(); ++outer) {
        swarm.predatorForces[outer] = PartialForce(); // identity object, useful so that we can just sum in the inner loop

        for (std::size_t relevantEntries = 0; relevantEntries < pforces.size(); relevantEntries += swarm.predatorForces.size()) {
            swarm.predatorForces[outer] += pforces[relevantEntries];
        }
    }

    // Finally, complete the force calculations...
    swarm.finalizePredatorForces();
    // and update the positions of the predators.
    swarm.updatePositions(swarm.predators, configuration.predator.limit.speed);
}

void Simulation::serialize()
{
    PartialSwarm &currentPart = swarm.parts[rank];

    // Write the boids ordered by rank into the file
    MPI_File_write_ordered(file, currentPart.boids.data(), currentPart.boids.size(),
                           MPI_BOID_THIN, MPI_STATUS_IGNORE);

    // Root also writes down the predators.
    // We could use multiple processes to store them, similar to
    // the way we used MPI_File_write_ordered above, but this would complicate
    // things and the performance impact might not even be positive, since, generally speaking,
    // there are way less predators than there are boids.
    if (isRoot()) {
        MPI_File_write_shared(file, swarm.predators.data(), swarm.predators.size(),
                              MPI_BOID_THIN, MPI_STATUS_IGNORE);
    }

    // Hard sync if there were predators to avoid a race condition
    if (!swarm.predators.empty()) {
        MPI_Barrier(MPI_COMM_WORLD);
    }
}
