// Swarm.cpp
// Defines a swarm of boids ('flock') that is broken into multiple divisions and has its
// behaviour calculated on multiple processors via MPI
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Swarm.h"
#include "PartialSwarm.h"
#include "../common/Flocking.h"
#include "../common/Configuration.h"

#include <cassert>
#include <cmath> // nextafter
#include <new> // placement new
#include <cstdint> // CHAR_BIT

// Construct a new Swarm with a given count of boids, predators and partial swarms
Swarm::Swarm(std::size_t boidCount, std::size_t predatorCount, int partialSwarmCount) :
    maxRank{partialSwarmCount}
{
    // Helper to create a boid with random position and -velocity in our world
    static auto randomBoid = []() -> Boid {
        static auto random = [](float range) {
            static std::once_flag once;
            std::call_once(once, []() {
                std::srand(configuration.deterministic ? 0 : std::time(nullptr));
            });

            return std::rand() / static_cast<float>(RAND_MAX) * range;
        };

        auto ws = configuration.worldSize;
        auto ls = configuration.boid.limit.speed;
        return {
            { random(ws), random(ws), random(ws) },
            { random(ls), random(ls), random(ls) }
        };
    };

    // Generate the boids
    boids.reserve(boidCount);
    std::generate_n(std::back_inserter(boids), boidCount, randomBoid);

    // Generate the predators
    predators.reserve(predatorCount);
    std::generate_n(std::back_inserter(predators), predatorCount, randomBoid);

    predatorForces.reserve(predatorCount);

    // Generate the partial swarms
    parts = reinterpret_cast<PartialSwarm *>(
        new char[sizeof(PartialSwarm) * partialSwarmCount]);
    for (int index = 0; index < partialSwarmCount; ++index) {
        new(parts + index) PartialSwarm(*this, index);
    }
}

// Free ressources
Swarm::~Swarm()
{
    // for (int index = 0; index < maxRank; ++index)
    //     (reinterpret_cast<PartialSwarm *>(parts) + index)->~PartialSwarm();
    // delete[] parts;
}

// Updates the partial swarm with the index 'rank' to include all boids that came over
// from neighbored partial swarms since the last call to this function
void Swarm::updatePartialSwarms(std::size_t rank)
{
    std::vector<std::vector<Boid>> buffer(maxRank);

    for (const auto &outer : parts[rank]) {
        for (const auto &inner : parts[rank]) {
            for (auto &boid : inner.boids) {
                if (boid.position.x >= outer.minX
                 && boid.position.x <  outer.maxX) {
                    buffer[outer.tag()].emplace_back(boid);
                }
            }
        }
    }

    for (auto &neighbor : parts[rank]) {
        neighbor.boids = std::move(buffer[neighbor.tag()]);
    }
}

// Check whether there is a predator in range and if so, avoid that predator
bool Swarm::avoidPredators(Boid &boid)
{
    for (const Predator &predator : predators) {
        // Vector from predator to boid
        Vector3f difference = calculateDifference(predator.position, boid.position);
        float    distance   = difference.length();

        // checks whether the predator chases the boid in question
        if (distance < (configuration.predator.range.cohesion / 2.0f)) {
            // evade as fast as possible
            boid.force = difference.normalize();
            boid.force *= configuration.boid.limit.speed;
            return true;
        }
    }

    return false;
}

// Calculate the steer towards other boids in range
// This method updates the boid's force field, so that these
// force vectors can be applied atomically
void Swarm::calculateSteer(Boid &boid, std::size_t rank)
{
    PartialForce separationSteer;
    PartialForce alignmentSteer;
    PartialForce cohesionSteer;

    for (const auto &neighbor : parts[rank]) {
        for (const auto &other : neighbor.boids) {
            float distance = calculateDifference(boid.position, other.position).length();

            if (distance > 0.0f && distance < configuration.boid.range.separation) {
                separationSteer += separation(boid, other);
            }

            if (distance > 0.0f && distance < configuration.boid.range.alignment) {
                alignmentSteer += alignment(boid, other);
            }

            if (distance > 0.0f && distance < configuration.boid.range.cohesion) {
                cohesionSteer += cohesion(boid, other);
            }
        }
    }

    boid.force = separationSteer.finalize(boid, PartialForce::Actor::Boid, PartialForce::Force::Separation)
               + alignmentSteer.finalize(boid, PartialForce::Actor::Boid, PartialForce::Force::Alignment)
               + cohesionSteer.finalize(boid, PartialForce::Actor::Boid, PartialForce::Force::Cohesion);
}

void Swarm::calculatePartialPredatorForce(Predator &predator, std::size_t rank)
{
    // only cohesionSteer is necessary, separation between predators can be computed later and
    // does not need to be computed part for part.
    PartialForce cohesionSteer;

    for (Boid &other : parts[rank].boids) {
        float distance = calculateDifference(predator.position, other.position).length();

        if (distance < configuration.predator.range.cohesion) {
            cohesionSteer += cohesion(predator, other);
        }
    }

    // index into original predator vector, used to index into
    // predatorForces vector
    unsigned int index = &predator - &predators[0];
    assert(index >= 0 && index < predators.size());

    predatorForces[index] = cohesionSteer;
}

void Swarm::finalizePredatorForces()
{
    // This operation assumes the partial forces have already been processed.

    // Calculate the separation between different predators
    for (std::size_t i = 0; i < predators.size(); ++i) {
        Predator &predator = predators[i];

        PartialForce separationSteer;

        for (const Predator &other : predators) {
            float distance = calculateDifference(predator.position, other.position).length();

            if (distance > 0.0f && distance < configuration.predator.range.separation) {
                separationSteer += separation(predator, other);
            }
        }

        // predatorForces should contain the sum of all partial forces from all other processes,
        // then the finalize method calculates the resulting vector.
        // This vector is only close to the 'exact' one, since floating point arithmetic is a mess.
        predator.force = predatorForces[i].finalize(predator, PartialForce::Actor::Predator, PartialForce::Force::Cohesion)
                       + separationSteer.finalize(predator, PartialForce::Actor::Predator, PartialForce::Force::Separation);
    }
}

// Calculate the predator attack behavior
void Swarm::calculatePredatorMovement(Predator &predator)
{
    PartialForce cohesionSteer;     // predators feel cohesion towards boids
    PartialForce separationSteer;   // and feel separation towards other predators

    for (Boid &other : boids) {
        float distance = calculateDifference(predator.position, other.position).length();

        if (distance < configuration.predator.range.cohesion) {
            // then go after this boid
            cohesionSteer += cohesion(predator, other);
        }
    }

    // separation towards other boids
    for (Predator &other : predators) {
        float distance = calculateDifference(predator.position, other.position).length();

        if (distance > 0.0f && distance < configuration.predator.range.separation) {
            separationSteer += separation(predator, other);
        }
    }

    predator.force = cohesionSteer.finalize(predator, PartialForce::Actor::Predator, PartialForce::Force::Cohesion)
                   + separationSteer.finalize(predator, PartialForce::Actor::Predator, PartialForce::Force::Separation);
}

// Update the boids/predators
// That is, calculate the new position
void Swarm::updatePositions(std::vector<Boid> &boids, const float speedLimit)
{
    // for every boid in current partial swarm
    for (auto &boid : boids) {
        // update velocity
        boid.velocity += boid.force;
        boid.velocity  = boid.velocity.limitLength(speedLimit);

        // update position
        boid.position += boid.velocity;

        // limit position to world border
        auto limit = [](float &value, float min, float max) {
            if (value < min) {
                value = max - std::abs(min - value);

                // Workaround for floating point arithmetic
                if (value == max) {
                    value = std::nextafter(value, min);
                }
            }
            else if (value >= max) {
                value = min + std::abs(max - value);
            }
        };

        limit(boid.position.x, 0.0f, configuration.worldSize);
        limit(boid.position.y, 0.0f, configuration.worldSize);
        limit(boid.position.z, 0.0f, configuration.worldSize);
    }
}
