// Swarm.hpp
// Defines a swarm of boids ('flock') that is broken into multiple divisions and has its
// behaviour calculated on multiple processors via MPI
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef SWARMFLOCKING_SWARM_H
#define SWARMFLOCKING_SWARM_H

#include <utility> // forward
#include <cstdint> // size_t
#include <cstdlib> // srand, rand
#include <iterator> // back_inserter
#include <vector> // vector
#include <mutex> // once_flag, call_once
#include <algorithm> // generate_n, copy_if

#include "../common/Boid.h"
#include "PartialForce.h"

class PartialSwarm;

class Swarm
{
public:
    // Construct a new Swarm with a given count of boids, predators and partial swarms
    Swarm(std::size_t boidCount, std::size_t predatorCount, int partialSwarmCount);

    // Free ressources
    ~Swarm();

    // Updates the partial swarm with the index rank to include all boids that came over
    // from neighbored partial swarms since the last call to this function
    void updatePartialSwarms(std::size_t rank);

    // Check whether there is a predator a range and avoid any
    bool avoidPredators(Boid &boid);

    // Calculate the steer towards other boids in range
    void calculateSteer(Boid &boid, std::size_t rank);

    void calculatePartialPredatorForce(Predator &predator, std::size_t rank);
    void finalizePredatorForces();

    // Calculate the predator attack behavior
    void calculatePredatorMovement(Predator &predator);

    // Update the boids/predators
    // That is, calculate the new position
    void updatePositions(std::vector<Boid> &boids, float speedLimit);

//private:
    PartialSwarm *parts;
    int maxRank;

    // Total Swarm
    std::vector<Boid> boids;

    // Predators
    std::vector<PartialForce> predatorForces;
    std::vector<Predator> predators;
};

#endif /* SWARMFLOCKING_SWARM_H */
