// Timer.h
// Simple timer to measure performance
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef SWARMFLOCKING_SIMULATION_TIMER_H
#define SWARMFLOCKING_SIMULATION_TIMER_H

#include <utility>

// -----------------------------------------------------------------------------
#if defined(__APPLE__)
// -----------------------------------------------------------------------------

#include <mach/mach_time.h>
#include <cstdint>

class Timer
{
public:
    // create a new Timer and start the internal clock
    inline Timer()
    {
        mach_timebase_info_data_t timebase;
        mach_timebase_info(&timebase);

        machFactor = static_cast<double>(timebase.numer) /
                     static_cast<double>(timebase.denom);

        absoluteTime = mach_absolute_time();
    }

    // restarts the internal clock and returns the number
    // of nanoseconds passed since the last invocation
    inline long restart()
    {
        long previousTime = absoluteTime;
        absoluteTime = mach_absolute_time();

        return static_cast<long>(machFactor * (absoluteTime - previousTime));
    }

private:
    std::uint64_t absoluteTime;
    double machFactor;
};

// -----------------------------------------------------------------------------
#elif defined(__linux__)
// -----------------------------------------------------------------------------

#include <ctime>
#include <cassert>

class Timer
{
public:
    // create a new Timer and start the internal clock
    inline Timer()
    {
        int ret;
    #ifdef CLOCK_MONOTONIC
        ret = clock_gettime(CLOCK_MONOTONIC, &ts);
    #else
        ret = clock_gettime(CLOCK_REALTIME, &ts);
    #endif
        assert(ret == 0);
    }

    // restarts the internal clock and returns the number
    // of nanoseconds passed since the last invocation
    inline long restart()
    {
        long previous = ts.tv_sec * 1000000000 + ts.tv_nsec;
        int ret = clock_gettime(CLOCK_MONOTONIC, &ts);
        assert(ret == 0);

        return (ts.tv_sec * 1000000000 + ts.tv_nsec) - previous;
    }

private:
    timespec ts;
};

// -----------------------------------------------------------------------------
#else
// -----------------------------------------------------------------------------

// TODO: Support windows (?)

#endif

template <typename Fn, typename... Args>
long timedCall(Fn&& f, Args&&... args)
{
    Timer t;
    std::forward<Fn>(f)(std::forward<Args>(args)...);
    return t.restart();
}

#endif /* SWARMFLOCKING_SIMULATION_TIMER_H */
