// main.cpp
// Defines the applications entry-point
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

// stdlibc++
#include <cstdlib>
#include <string>
#include <fstream>

// MPI
#include <mpi.h>

// boost libraries
#include <boost/program_options.hpp>

// SwarmFlocking
#include "Simulation.h"


int main(int argc, char *argv[])
{
	// Initialize MPI
	MPI_Init(&argc, &argv);

    // Command line argument buffers
    std::size_t boidCount = 1024;
	std::size_t predatorCount = 0;
    std::size_t steps = 0;
    std::string output;

	// Setup boost.program_options for
	using namespace boost::program_options;
    options_description description{"Options"};
    description.add_options()
        ("help,h", "Print this help message")
        ("boid-count,b", value<std::size_t>(&boidCount)->required(), "Number of boids to simulate")
        ("predator-count,p", value<std::size_t>(&predatorCount)->required(), "Number of predators to simulate")
        ("steps,s", value<std::size_t>(&steps)->required(), "Number of steps to simulate")
        ("output,o", value<std::string>(&output)->required(), "Specify an output file");
        // add option for no output, detailed stats after running.
        // TODO stdout (overwrite --output)

	// Parse the command line options
    variables_map variablesMap;
    try {
        store(parse_command_line(argc, argv, description), variablesMap);

        if (variablesMap.count("help")) {
            std::cout << description << std::endl; // print usage
            
            MPI_Finalize();
            return EXIT_SUCCESS;
        }

        notify(variablesMap);
    } catch (const error &err) {
        std::cerr << "Erroneous usage of command line parameter: "
                  << err.what() << std::endl << std::endl
                  << description << std::endl;

        MPI_Finalize();
        return EXIT_FAILURE;
    }

	// Try opening the output file
	MPI_File file;
	if (MPI_SUCCESS != MPI_File_open(MPI_COMM_WORLD, const_cast<char *>(output.c_str()),
				MPI_MODE_CREATE | MPI_MODE_WRONLY, MPI_INFO_NULL, &file)) {
		std::cerr << "Failed to open output file. Make sure you have sufficient "
				  << "privileges to work!" << std::endl;

		MPI_Finalize();
        return EXIT_FAILURE;
	}

	// Total and current process
	int maxRank;
	int rank;
	MPI_Comm_size(MPI_COMM_WORLD, &maxRank);
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);

	// Create and run the simulation
    Simulation simulation{boidCount, predatorCount, file, steps, maxRank, rank};
    simulation.run();

	// Free MPI ressources
    MPI_File_close(&file);
	MPI_Finalize();

    return EXIT_SUCCESS;
}
