// BoidBuffer.cpp
// Handles deserialisation of boid data from input streams.
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "BoidBuffer.h"
#include <future>     // for std::async, std::await

static std::future<BoidBuffer::SwarmBuf> res;

BoidBuffer::BoidBuffer(std::istream &in, const SwarmStorageConfiguration &configuration) :
    input{in},
    bufferEmpty{false},
    config(configuration)
{
    boidCount       = config.boidCount;
    predatorCount   = config.predatorCount;

    boidA = new Boid[boidCount];
    boidB = new Boid[boidCount];

    predA = new Boid[predatorCount];
    predB = new Boid[predatorCount];

    // Begin fetching the first block of data
    res = std::async(std::launch::async,
                     &BoidBuffer::requestNext,
                     this, boidA, predA);
}

BoidBuffer::~BoidBuffer()
{
    delete[] boidA;
    delete[] boidB;

    delete[] predA;
    delete[] predB;
}

/*
    requestNext reads the next boid frame from the input stream.
    requestNext returns a SwarmBuf structure, where the 'isValid' field
    reflects any potential error.
*/
BoidBuffer::SwarmBuf BoidBuffer::requestNext(Boid *intoBoid, Predator *intoPred)
{
    for (std::size_t i = 0; i < config.boidCount; ++i) {
        intoBoid[i].deserialize(input, config.positionOnly);
    }
    for (std::size_t i = 0; i < config.predatorCount; ++i) {
        intoPred[i].deserialize(input, config.positionOnly);
    }

    BoidBuffer::SwarmBuf ret = {
        boidCount,
        predatorCount,
        intoBoid,
        intoPred,
        !(input.bad() || input.eof() || input.fail())
    };

    return ret;
}

/*
    Returns a SwarmBuf structure that holds the results.
    If there is no more data or there was an error, the field
    'isValid' is set to false.
    This method also starts a new thread to asynchronously
    fetch more data
*/
BoidBuffer::SwarmBuf BoidBuffer::next()
{
    SwarmBuf ret;

    if (bufferEmpty || !(ret = res.get()).isValid) {
        bufferEmpty = true;
        return ret;
    } else {
        // start fetching of the next batch of boid data
        res = std::async(std::launch::async,
                         &BoidBuffer::requestNext,
                         this, (ret.boids == boidA) ? boidB : boidA, (ret.predators == predA) ? predB : predA);
        // return ptr to previously asynchronously fetched data
        return ret;
    }
}
