// Visualisation.cpp
// Manages visual presentation of data using SFML.
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Visualisation.h"
#include "BoidClustering.h"

#include <SFML/System.hpp>
#include <SFML/Window.hpp>
#include <SFML/OpenGL.hpp>

#include <cstdint> // uint32_t
#include <mutex> // std::call_once

#include "../common/Configuration.h"

Visualisation::Visualisation(sf::String title,
                             std::istream &inputSource,
                             const SwarmStorageConfiguration &configuration,
                             bool singleStepping,
                             std::size_t fps) :
    window{sf::VideoMode{800, 800, 24}, title, sf::Style::Default},
    font{},
    dataSource{inputSource, configuration},
    singleStepping{singleStepping},
    nextFrameRequested{true},
    running{true}
{
    window.setFramerateLimit(fps);

    window.setVerticalSyncEnabled(false);
    window.setActive(true);
    window.resetGLStates();

    glEnable(GL_MULTISAMPLE);
    glEnable(GL_LINE_SMOOTH);
}

void Visualisation::run()
{
    BoidBuffer::SwarmBuf buffer;
    while (window.isOpen()) {
        if (    (singleStepping  && nextFrameRequested)
            ||  (!singleStepping && running)) {

            BoidBuffer::SwarmBuf oldBuf = buffer;
            buffer = dataSource.next();

            if (!buffer.isValid) {
                 buffer = oldBuf; // continue displaying something after input is consumed
            } else {
                // update clusters
                static std::once_flag once;
                std::call_once(once, [this, &buffer] {
                    clusters = std::vector<int>(buffer.nboids);
                });

                if (singleStepping)
                    kmeans(buffer.boids, buffer.nboids, clusters, 200);
             }

            nextFrameRequested = false;
        }

        sf::Event event;
        if (window.pollEvent(event)) {
            switch (event.type) {
                case sf::Event::Closed:
                    window.close();
                    break;

                case sf::Event::Resized:
                    handleSizeEvent(event.size);
                    break;

                case sf::Event::KeyPressed:
                    handleKeyEvent(event.key);
                    break;

                case sf::Event::TextEntered:
                    handleTextEvent(event.text);
                    break;

                case sf::Event::MouseMoved:
                    handleMouseMoveEvent(event.mouseMove);
                    break;

                case sf::Event::MouseButtonPressed:
                    handleMouseButtonEvent(event.mouseButton);

                case sf::Event::MouseWheelMoved:
                    handleMouseWheelEvent(event.mouseWheel);

                default:
                    break;
            }
        }

        glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        sf::Vector2u wsize = window.getSize();
        glViewport(0, 0, wsize.x, wsize.y);
        gluPerspective(60, (float)wsize.x / (float)wsize.y, 0.1f, 256.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glPushMatrix();
        glTranslatef(0.0f, 0.0f, -3.5f);

        glRotatef(rotation[0], 1.f, 0.f, 0.f);
        glRotatef(rotation[1], 0.f, 1.f, 0.f);

        renderCube();
        renderSwarm(buffer);

        glPopMatrix();

        window.pushGLStates();

        window.display();
        window.popGLStates();
    }
}

void Visualisation::handleSizeEvent(sf::Event::SizeEvent sizeEvent)
{
    glViewport(0, 0, sizeEvent.width, sizeEvent.height);
}

void Visualisation::handleKeyEvent(sf::Event::KeyEvent keyEvent)
{
    switch (keyEvent.code) {
        case sf::Keyboard::Escape:
            window.close();
            break;

        case sf::Keyboard::Num1:
            rotationSpeed += 0.5f;
            break;

        case sf::Keyboard::Num2:
            if ((rotationSpeed -= 0.5f) < 0.0f) {
                rotationSpeed = 0.5f;
            }
            break;

        case sf::Keyboard::Up:
            rotation[0] -= rotationSpeed;
            break;

        case sf::Keyboard::Down:
            rotation[0] += rotationSpeed;
            break;

        case sf::Keyboard::Left:
            rotation[1] -= rotationSpeed;
            break;

        case sf::Keyboard::Right:
            rotation[1] += rotationSpeed;
            break;

        case sf::Keyboard::Space:
            if (singleStepping)
                nextFrameRequested = true;
            else
                running = !running;
            break;

        default:
            break;
    }
}

void Visualisation::handleTextEvent(sf::Event::TextEvent textEvent)
{

}

void Visualisation::handleMouseMoveEvent(sf::Event::MouseMoveEvent mouseMoveEvent)
{

}

void Visualisation::handleMouseButtonEvent(sf::Event::MouseButtonEvent mouseButtonEvent)
{

}

void Visualisation::handleMouseWheelEvent(sf::Event::MouseWheelEvent mouseWheelEvent)
{

}

void Visualisation::renderSwarm(BoidBuffer::SwarmBuf &buffer)
{
     constexpr uint32_t indexColors[] = {
        0x000000, 0xFFFF00, 0x1CE6FF, 0xFF34FF, 0xFF4A46, 0x008941, 0x006FA6, 0xA30059,
        0xFFDBE5, 0x7A4900, 0x0000A6, 0x63FFAC, 0xB79762, 0x004D43, 0x8FB0FF, 0x997D87,
        0x5A0007, 0x809693, 0xFEFFE6, 0x1B4400, 0x4FC601, 0x3B5DFF, 0x4A3B53, 0xFF2F80,
        0x61615A, 0xBA0900, 0x6B7900, 0x00C2A0, 0xFFAA92, 0xFF90C9, 0xB903AA, 0xD16100,
        0xDDEFFF, 0x000035, 0x7B4F4B, 0xA1C299, 0x300018, 0x0AA6D8, 0x013349, 0x00846F,
        0x372101, 0xFFB500, 0xC2FFED, 0xA079BF, 0xCC0744, 0xC0B9B2, 0xC2FF99, 0x001E09,
        0x00489C, 0x6F0062, 0x0CBD66, 0xEEC3FF, 0x456D75, 0xB77B68, 0x7A87A1, 0x788D66,
        0x885578, 0xFAD09F, 0xFF8A9A, 0xD157A0, 0xBEC459, 0x456648, 0x0086ED, 0x886F4C,
        0x34362D, 0xB4A8BD, 0x00A6AA, 0x452C2C, 0x636375, 0xA3C8C9, 0xFF913F, 0x938A81,
        0x575329, 0x00FECF, 0xB05B6F, 0x8CD0FF, 0x3B9700, 0x04F757, 0xC8A1A1, 0x1E6E00,
        0x7900D7, 0xA77500, 0x6367A9, 0xA05837, 0x6B002C, 0x772600, 0xD790FF, 0x9B9700,
        0x549E79, 0xFFF69F, 0x201625, 0x72418F, 0xBC23FF, 0x99ADC0, 0x3A2465, 0x922329,
        0x5B4534, 0xFDE8DC, 0x404E55, 0x0089A3, 0xCB7E98, 0xA4E804, 0x324E72, 0x6A3A4C,
        0x83AB58, 0x001C1E, 0xD1F7CE, 0x004B28, 0xC8D0F6, 0xA3A489, 0x806C66, 0x222800,
        0xBF5650, 0xE83000, 0x66796D, 0xDA007C, 0xFF1A59, 0x8ADBB4, 0x1E0200, 0x5B4E51,
        0xC895C5, 0x320033, 0xFF6832, 0x66E1D3, 0xCFCDAC, 0xD0AC94, 0x7ED379, 0x012C58
    };

    // static std::vector<int> clusters(buffer.nboids);
    // kmeans(buffer.boids, buffer.nboids, clusters, 100);

    float zoom[3] = {
        configuration.worldSize / 2.0f,
        configuration.worldSize / 2.0f,
        configuration.worldSize / 2.0f
    };

    glColor3f(0.0f, 0.0f, 1.0f);
    glPointSize(2.0f);

    glBegin(GL_POINTS);

        for (std::size_t i = 0; i < buffer.nboids; ++i) {
            Boid &current = buffer.boids[i];

            uint32_t color = indexColors[clusters[i]];

            if (singleStepping) {
                glColor3f(static_cast<float>(color & 0x0000ff) / 255,
                          static_cast<float>((color & 0x00ff00) >> 8) / 255,
                          static_cast<float>((color & 0xff0000) >> 16) / 255);
            } else {
                glColor3f(0.0f, 0.0f, 1.0f);
            }

            glVertex3f((current.position.x - zoom[0]) / zoom[0],
                       (current.position.y - zoom[1]) / zoom[1],
                       (current.position.z - zoom[2]) / zoom[2]);
        }

    glEnd();


    glColor3f(1.0f,0.0f,0.0f); // red color
    glPointSize(4.0);

    glBegin(GL_POINTS);

        for (std::size_t i = 0; i < buffer.npredators; ++i) {
            Predator &predator = buffer.predators[i];

            glVertex3f((predator.position.x - zoom[0]) / zoom[0],
                       (predator.position.y - zoom[1]) / zoom[1],
                       (predator.position.z - zoom[2]) / zoom[2]);
        }

    glEnd();
}

void Visualisation::renderCube()
{
    static constexpr float cubeVertices[8][3] = {
        { -1.0, -1.0,  1.0},
        { -1.0,  1.0,  1.0},
        {  1.0,  1.0,  1.0},
        {  1.0, -1.0,  1.0},
        { -1.0, -1.0, -1.0},
        { -1.0,  1.0, -1.0},
        {  1.0,  1.0, -1.0},
        {  1.0, -1.0, -1.0},
    };

    glBegin(GL_LINES);      // Draw The Cube Using lines

    glColor3f(0.0f, 0.0f, 0.0f);

    // top
    glVertex3fv(cubeVertices[0]);
    glVertex3fv(cubeVertices[1]);

    glVertex3fv(cubeVertices[1]);
    glVertex3fv(cubeVertices[2]);

    glVertex3fv(cubeVertices[2]);
    glVertex3fv(cubeVertices[3]);

    glVertex3fv(cubeVertices[3]);
    glVertex3fv(cubeVertices[0]);

    // bottom
    glVertex3fv(cubeVertices[4]);
    glVertex3fv(cubeVertices[5]);

    glVertex3fv(cubeVertices[5]);
    glVertex3fv(cubeVertices[6]);

    glVertex3fv(cubeVertices[6]);
    glVertex3fv(cubeVertices[7]);

    glVertex3fv(cubeVertices[7]);
    glVertex3fv(cubeVertices[4]);

    // connections
    glVertex3fv(cubeVertices[0]);
    glVertex3fv(cubeVertices[4]);

    glVertex3fv(cubeVertices[1]);
    glVertex3fv(cubeVertices[5]);

    glVertex3fv(cubeVertices[2]);
    glVertex3fv(cubeVertices[6]);

    glVertex3fv(cubeVertices[3]);
    glVertex3fv(cubeVertices[7]);

    glEnd(); // End Drawing The Cube
}
