// main.cpp
// Entry point for visualisation. Parses command line options and hands over control
// of execution to Visualisation.cpp
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <fstream>

#include <boost/program_options.hpp>

#include "Visualisation.h"
#include "../common/SwarmStorageConfiguration.h"

int main(int argc, char *argv[])
{
    std::ifstream inputStream;
    std::string   input;
    std::size_t   fps = 30;

    // lets user interactively skip frames
    // by hitting the space bar.
    bool          singleStepping = false;

    // useful for streaming large amounts of data directly
    // simulation binary to the visualisation
    bool          useStdin = false;

    std::size_t boidCount;
    std::size_t predatorCount;

    namespace po = boost::program_options;
    po::options_description desc("Options");
    desc.add_options()
        ("help,h", "Print this help message")
        ("input,i", po::value<std::string>(&input), "input file source")
        ("fps", po::value<std::size_t>(&fps), "Set a custom number of frames to be displayed each second (defaults to 30)")
        ("single-stepping,ss", "Control execution of the visualisation by pressing \'space\'")
        ("stdin", "use stdin as source (overwrites --input)")
        ("boid-count,b", po::value<std::size_t>(&boidCount)->required(), "Number of boids per frame")
        ("predator-count,p", po::value<std::size_t>(&predatorCount)->required(), "Number of predators per frame");

    po::variables_map vm;
    try {
        po::store(po::parse_command_line(argc, argv, desc), vm);

        if (vm.count("help")) {
            std::cout << desc << std::endl; // print usage
            return EXIT_SUCCESS;
        }

        if (!vm.count("input") && !vm.count("stdin")) {
            std::cerr << "Please specify one of the following options" << std::endl
                      << "\t --input" << std::endl
                      << "\t --stdin" << std::endl << std::endl;

            std::cout << desc << std::endl; // print usage
            return EXIT_FAILURE;
        }

        useStdin = vm.count("stdin");
        singleStepping = vm.count("single-stepping");

        po::notify(vm);
    } catch (po::error &e) {
        std::cerr << "Erroneous usage of command line parameter: " << e.what() << std::endl << std::endl;
        std::cout << desc << std::endl;

        return EXIT_FAILURE;
    }

    // if input file is specified
    if (input != "") {
        inputStream.open(input, std::ios::in | std::ios::binary);
        if (!inputStream.is_open() || inputStream.fail()) {
            std::cerr << "Failed to open specified input file" << std::endl << std::endl;
            return EXIT_FAILURE;
        }
    }

    SwarmStorageConfiguration config = { boidCount, predatorCount, true /* position only */ };

    Visualisation visualisation{"papo.SwarmFlocking",
                                useStdin ? std::cin : inputStream,
                                config,
                                singleStepping,
                                fps};
    visualisation.run();

    return EXIT_SUCCESS;
}
