// Vector.h
// Basic vector maths
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef SWARMFLOCKING_VECTOR_H
#define SWARMFLOCKING_VECTOR_H

#include <type_traits>
#include <utility>
#include <cmath>
#include <immintrin.h>

template <std::size_t N, typename T>
class Vector;

using Vector3f = Vector<3, float>;
using Vector3d = Vector<3, double>;

template <typename T>
class alignas(0x10) Vector<3, T>
{
    static_assert(std::is_floating_point<T>::value,
                  "Vector requires T to be a floating point type");

public:
    Vector() : x{T{}}, y{T{}}, z{T{}}, padding{T{}}
    {
    }

    Vector(T x, T y, T z)
        : x{x}, y{y}, z{z}, padding{T{}}
    {
    }

    Vector(const Vector &vec)
        : x{vec.x}, y{vec.y}, z{vec.z}, padding{T{}}
    {
    }

    Vector(Vector &&vec)
        : x{std::move(vec.x)}, y{std::move(vec.y)}, z{std::move(vec.z)}, padding{T{}}
    {
    }

    Vector &operator=(const Vector& rhs)
    {
        x = rhs.x;
        y = rhs.y;
        z = rhs.z;
        return *this;
    }

    Vector &operator=(Vector &&rhs)
    {
        x = std::move(rhs.x);
        y = std::move(rhs.y);
        z = std::move(rhs.z);
        return *this;
    }

    T lengthsq() const
    {
        return *this * *this;
        // return (x * x) + (y * y) + (z * z);
    }

    T length() const
    {
        return std::sqrt(lengthsq());
    }

    Vector setLength(T l) const
    {
        return normalize() * l;
    }

    Vector limitLength(T l) const
    {
        auto len = length();
        if (len > l)
            return setLength(l);
        else
            return *this;
    }

    Vector normalize() const
    {
        auto len = length();
        return *this / len;
        // return {x / len, y / len, z / len};
    }

    Vector operator+(const Vector &vec) const
    {
        Vector ret;

        __m128 a = _mm_load_ps((float *)this);
        __m128 b = _mm_load_ps((float *)&vec);

        _mm_store_ps((float *)&ret, _mm_add_ps(a, b));
        return ret;

        // return {x + vec.x, y + vec.y, z + vec.z};
    }

    Vector &operator+=(const Vector &vec)
    {
        __m128 a = _mm_load_ps((float *)this);
        __m128 b = _mm_load_ps((float *)&vec);

        _mm_store_ps((float *)this, _mm_add_ps(a, b));

        // x += vec.x;
        // y += vec.y;
        // z += vec.z;

        return *this;
    }

    Vector operator-(const Vector &vec) const
    {
        Vector ret;

        __m128 a = _mm_load_ps((float *)this);
        __m128 b = _mm_load_ps((float *)&vec);

        _mm_store_ps((float *)&ret, _mm_sub_ps(a, b));
        return ret;
        
        // return {x - vec.x, y - vec.y, z - vec.z};
    }

    Vector &operator-=(const Vector &vec)
    {
        __m128 a = _mm_load_ps((float *)this);
        __m128 b = _mm_load_ps((float *)&vec);

        _mm_store_ps((float *)this, _mm_sub_ps(a, b));

        // x -= vec.x;
        // y -= vec.y;
        // z -= vec.z;

        return *this;
    }

    T operator*(const Vector &vec) const
    {
        Vector3f out;
        __m128 t   = _mm_load_ps((float *)this);
        __m128 res = _mm_mul_ps(t, t);
        _mm_store_ps((float *)&out, res);
        
        return out.x + out.y + out.z;
        // return x * vec.x + y * vec.y + z * vec.z;
    }

    Vector operator*(T val) const
    {
        Vector3f ret;

        __m128 t = _mm_load_ps((float *)this);
        __m128 f = _mm_set1_ps(val);

        _mm_store_ps((float *)&ret, _mm_mul_ps(t, f));
        return ret;

        // return {x * val, y * val, z * val};
    }

    Vector &operator*=(T val)
    {
        __m128 t = _mm_load_ps((float *)this);
        __m128 f = _mm_set1_ps(val);

        _mm_store_ps((float *)this, _mm_mul_ps(t, f));
        
        // x *= val;
        // y *= val;
        // z *= val;
        
        return *this;
    }

    Vector operator/(T val) const
    {
        Vector3f ret;

        __m128 t = _mm_load_ps((float *)this);
        __m128 f = _mm_set1_ps(val);

        _mm_store_ps((float *)&ret, _mm_div_ps(t, f));
        return ret;

        // return {x / val, y / val, z / val};
    }

    Vector &operator/=(T val)
    {
        __m128 t = _mm_load_ps((float *)this);
        __m128 f = _mm_set1_ps(val);

        _mm_store_ps((float *)this, _mm_div_ps(t, f));
        
        // x /= val;
        // y /= val;
        // z /= val;
        
        return *this;
    }

    T x;
    T y;
    T z;
    T padding;
};

#endif /* SWARMFLOCKING_VECTOR_H */
