// Simulation.cpp
// Manages the simulation. Responsible for timing, serialisation and 
// synchronisation in the MPI variant.
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Simulation.h"

#include "Timer.h"
#include <cstddef>
#include <cassert>
#include <iostream>
#include <mpi.h>

static MPI_Aint type_extent(MPI_Datatype type)
{
    MPI_Aint lowerbound, extent;
    MPI_Type_get_extent(type, &lowerbound, &extent);
    return extent;
}

void Simulation::MPI_VECTOR3F_init()
{
    MPI_Datatype types[] = { MPI_FLOAT, MPI_FLOAT, MPI_FLOAT };
    int blocklengths[] = {1, 1, 1};
    const MPI_Aint displacements[] = { offsetof(Vector3f, x),
                                       offsetof(Vector3f, y),
                                       offsetof(Vector3f, z) };
    
    // temporary data type, missing the padding data
    MPI_Datatype _MPI_VECTOR3F;
    MPI_Type_create_struct(3, blocklengths, displacements, types, &_MPI_VECTOR3F);

    // resize the data type to account for .padding field (this does not get send!)
    MPI_Type_create_resized(_MPI_VECTOR3F, 0, sizeof(Vector3f), &MPI_VECTOR3F);
    MPI_Type_commit(&MPI_VECTOR3F);

    // Error checking and datatype validation
    // The actual extent should be as large as the underlying data...
    assert(type_extent(MPI_VECTOR3F) == sizeof(Vector3f));

    // but the padding field does not need to be sent
    int MPI_VECTOR3F_size;
    MPI_Type_size(MPI_VECTOR3F, &MPI_VECTOR3F_size);
    assert(MPI_VECTOR3F_size == (sizeof(float) * 3));
}

void Simulation::MPI_BOID_init()
{
    MPI_Datatype types[] = { MPI_VECTOR3F, MPI_VECTOR3F };
    int blocklengths[] = { 1, 1 };
    const MPI_Aint displacements[] = { offsetof(Boid, position),
                                       offsetof(Boid, velocity) };
    MPI_Type_create_struct(2, blocklengths, displacements, types, &MPI_BOID);
    MPI_Type_commit(&MPI_BOID);

    // Error checking and datatype validation
    assert(type_extent(MPI_BOID) == sizeof(Boid));
    
    int MPI_VECTOR3F_size, MPI_BOID_size;
    MPI_Type_size(MPI_VECTOR3F, &MPI_VECTOR3F_size);
    MPI_Type_size(MPI_BOID, &MPI_BOID_size);

    assert(MPI_BOID_size == (2 * MPI_VECTOR3F_size));
}

void Simulation::MPI_BOID_THIN_init()
{
    MPI_Datatype types[] = { MPI_VECTOR3F }; // just the position
    int blocklengths[] = { 1, 1 };
    const MPI_Aint displacements[] = { offsetof(Boid, position) };
    
    // temporary data type, missing velocity field!
    MPI_Datatype _MPI_BOID_THIN;
    MPI_Type_create_struct(1, blocklengths, displacements, types, &_MPI_BOID_THIN);

    MPI_Type_create_resized(_MPI_BOID_THIN, 0, sizeof(Boid), &MPI_BOID_THIN);
    MPI_Type_commit(&MPI_BOID_THIN);

    // Error checking and datatype validation
    assert(type_extent(MPI_BOID_THIN) == sizeof(Boid));

    int MPI_VECTOR3F_size, MPI_BOID_THIN_size;
    MPI_Type_size(MPI_VECTOR3F, &MPI_VECTOR3F_size);
    MPI_Type_size(MPI_BOID_THIN, &MPI_BOID_THIN_size);

    assert(MPI_VECTOR3F_size == MPI_BOID_THIN_size);
}

void Simulation::MPI_Datatypes_init()
{
    MPI_VECTOR3F_init();
    MPI_BOID_init();
    MPI_BOID_THIN_init();
}

// create a new simulation with a given amount of boids
Simulation::Simulation(std::size_t boidCount, 
                       std::size_t predatorCount,
                       MPI_File file,
                       std::size_t steps) : 
    swarm{boidCount, predatorCount},
    steps{steps},
    fh{file}
{
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    
    std::size_t nboidsPerProc, nboidsRem;
    nboidsPerProc = boidCount / size;
    nboidsRem     = boidCount % size;

    int beginnings[size + 1];
    beginnings[size] = boidCount;

    for (std::size_t proc = 0, from = 0; proc < size; ++proc) {
        beginnings[proc] = from;
        from += nboidsPerProc + ((proc < nboidsRem) ? 1 : 0);
    }

    from = beginnings[rank];
    to   = beginnings[rank+1];

    recvcounts = new int[size];
    displs     = new int[size];

    // correctly set up recvcount and displs for MPI communication
    for (std::size_t i = 0; i < size; ++i) {
        recvcounts[i] = beginnings[i+1] - beginnings[i];
        displs[i]     = beginnings[i];
    }

    MPI_Datatypes_init();
}

Simulation::~Simulation()
{
    delete[] recvcounts;
    delete[] displs;
}

// main loop
void Simulation::run()
{
    Timer timer;

    std::size_t nframes = 0;

    while (true) {
        step();
        ++nframes;

        if ((steps != 0) && (nframes == steps)) {
            return;
        }

        if (isRoot()) {
             auto elapsedTime = timer.restart();
             float framerate = 1e9 / elapsedTime;

             std::cerr << "Current FPS: " << std::to_string(framerate) << " Hz\r";
        }
    }
}

// runs a single step of swarm behaviour
void Simulation::step()
{
    startSynchronization();
    startSerialization();

    auto finishSync = [this]() {
        MPI_Wait(&req, MPI_STATUS_IGNORE);
    };

    auto finishSerialisation = [this]() {
        MPI_File_write_ordered_end(fh, &swarm.boids[from], MPI_STATUS_IGNORE);
    };

    swarm.update(from, to, finishSync, finishSerialisation);
}

// synchronizes the swarm
void Simulation::startSynchronization()
{
    MPI_Iallgatherv(MPI_IN_PLACE, 0, MPI_DATATYPE_NULL,
                    (void *)&swarm.boids[0], recvcounts, displs, MPI_BOID, 
                    MPI_COMM_WORLD, &req);
}

// write data for current step to disk!
void Simulation::startSerialization()
{
    MPI_File_write_ordered_begin(fh, &swarm.boids[from], to - from /* recvcounts[rank] */, MPI_BOID_THIN);
}

bool Simulation::isRoot() 
{
    return rank == 0;
}
