// Simulation.h
//
// Manages the simulation. Responsible for timing, serialisation and 
// synchronisation in the MPI variant.
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef SWARMFLOCKING_SIMULATION_H
#define SWARMFLOCKING_SIMULATION_H

#include <cstdint> // size_t
#include <mpi.h>
#include "Swarm.h"
#include "../common/SwarmStorageConfiguration.h"

class Simulation
{
public:
    // create a new simulation with a given amount of boids,
    // given number of predators and the output stream to log to.
    explicit Simulation(std::size_t boidCount, 
                        std::size_t predatorCount,
                        MPI_File file, 
                        std::size_t steps = 0);
    ~Simulation();

    // main loop
    void run();

private:
    // runs a single step of swarm behaviour
    void step();

    void startSynchronization();
    void startSerialization();

    bool isRoot();

    // the simulated swarm
    Swarm swarm;

    const std::size_t steps;

    // mpi data
    int size, rank;

    std::size_t from, to;

    SwarmStorageConfiguration storage_config;

    /* MPI related stuff follows

       Datatypes, Request of objects for async communications and File objects
       for serialisation */

    MPI_Datatype MPI_VECTOR3F;
    void MPI_VECTOR3F_init();

    MPI_Datatype MPI_BOID;
    void MPI_BOID_init();

    // special datatype that only stores / transmit the position 
    // of a boid. This is used for more efficient serialisation
    MPI_Datatype MPI_BOID_THIN;
    void MPI_BOID_THIN_init();

    // initialize all MPI datatypes used by this application
    void MPI_Datatypes_init();

    // request object for asynchronous communication
    MPI_Request req;

    int * recvcounts, // number of elements to receive from process i 
        * displs;     // offset from beginning of buffer for process i

    // fh for output
    MPI_File fh;
};

#endif /* SWARMFLOCKING_SIMULATION_H */
