// Swarm.cpp
// Defines a swarm of boids ('flock') that is broken into multiple divisions and has its
// behaviour calculated on multiple processors via MPI
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Swarm.h"

#include <cstdlib> // rand, srand
#include <ctime> // time
#include <cmath> // abs
#include <algorithm> // min
#include <cassert> // assert
#include <cstring> // memcpy

#include <mutex> // call_once

#include "Flocking.h"
#include "../common/Configuration.h"

// construct a new swarm with a given number of boids
// and a given number of predators
Swarm::Swarm(std::size_t boidCount, std::size_t predatorCount) :
    boids{boidCount},
    boidForces{boidCount},
    boidCount{boidCount},
    predators{predatorCount},
    predatorForces{predatorCount},
    predatorCount{predatorCount}
{
    // basic rng function
    static auto random = [](float range) {
        // deterministic for now, to better test this stuff

        // static std::once_flag once;
        // std::call_once(once, []() { std::srand(std::time(nullptr)); });

        return std::rand() / static_cast<float>(RAND_MAX) * range;
    };

    // set swarm positions and velocities
    for (Boid& boid: boids) {
        boid.position = {
            random(configuration.world.x),
            random(configuration.world.y),
            random(configuration.world.z)
        };

        boid.velocity = {
            random(configuration.boid.limit.speed),
            random(configuration.boid.limit.speed),
            random(configuration.boid.limit.speed)
        };
    }

    // set predator positions and velocities
    for (Predator& predator : predators) {
        predator.position = {
            random(configuration.world.x),
            random(configuration.world.y),
            random(configuration.world.z)
        };

        predator.velocity = {
            random(configuration.predator.limit.speed),
            random(configuration.predator.limit.speed),
            random(configuration.predator.limit.speed)
        };
    }
}

// construct a new swarm with a given number of boids
Swarm::Swarm(std::size_t boidCount) :
    Swarm{boidCount, 1}
{
}

// update the whole swarm
void Swarm::update()
{
    update(0, boids.size());
}

// update a given part of the swarm
// this does not work in an MPI environment with
// async communication!
void Swarm::update(std::size_t first, std::size_t last)
{
    // calculate forces for boids
    #pragma omp parallel for
    for (std::size_t index = first; index < last; ++index) {
        if (!avoidPredators(index)) {
            calculateSteer(index);
        }
    }

    // calculate forces for predators
    #pragma omp parallel for
    for (std::size_t index = 0; index < predators.size(); ++index) {
        calculatePredatorMovement(index);
    }

    // ... and apply those forces
    updateBoids(first, last);
    updatePredators();
}

// returns whether the boid has to avoid any predator
bool Swarm::avoidPredators(std::size_t index)
{
    for (const Predator& p : predators) {
        // Vector from predator to boid
        // Vector3f difference = boids[index].position - p.position;
        Vector3f difference = calculateDifference(p.position, boids[index].position);
        float    distance   = difference.length();

        // checks whether the predator chases the boid in question
        if (distance < (configuration.predator.range.cohesion / 2.0f)) {
            // evade as fast as possible
            boidForces[index] = difference.normalize();
            boidForces[index] *= configuration.boid.limit.speed;
            return true;
        }
    }

    return false;
}

void Swarm::calculatePartialSteer(std::size_t index, 
                                  std::size_t first, std::size_t last,
                                  partial_steer_t *out)
{
    Vector3f separationSteer, alignmentSteer, cohesionSteer;

    std::size_t separationInfluenceCount = 0,
                alignmentInfluenceCount  = 0,
                cohesionInfluenceCount   = 0;

    const Boid &b = boids[index];

    for (std::size_t i = first; i < last; ++i) {
        const Boid &other = boids[i];

        float distance = calculateDifference(b.position, other.position).length();

        if (distance > 0.0f && distance < configuration.boid.range.separation) {
            separationSteer += separation(b, other);
            ++separationInfluenceCount;
        }

        if (distance > 0.0f && distance < configuration.boid.range.alignment) {
            alignmentSteer += alignment(b, other);
            ++alignmentInfluenceCount;
        }

        if (distance > 0.0f && distance < configuration.boid.range.cohesion) {
            cohesionSteer += cohesion(b, other);
            ++cohesionInfluenceCount;
        }
    }

    partial_steer_t r = { 
        separationSteer, 
        alignmentSteer, 
        cohesionSteer,
        separationInfluenceCount, 
        alignmentInfluenceCount, 
        cohesionInfluenceCount,
        first, last, false /* never evades predator iff this gets run */
    };

    std::memcpy(out, &r, sizeof(partial_steer_t));
}

void Swarm::finishSteerCalculation(std::size_t index, 
                                   partial_steer_t *partial)
{
    Vector3f separationSteer = partial->separation, 
             alignmentSteer  = partial->alignment, 
             cohesionSteer   = partial->cohesion;

    std::size_t separationInfluenceCount = partial->separationInfluenceCount,
                alignmentInfluenceCount  = partial->alignmentInfluenceCount,
                cohesionInfluenceCount   = partial->cohesionInfluenceCount;

    const Boid &b = boids[index];

    for (std::size_t i = 0; i < boids.size(); ++i) {
        // skip the part we already covered
        if (i == partial->start) {
            i = partial->end - 1;
            continue;
        }

        const Boid &other = boids[i];

        float distance = calculateDifference(b.position, other.position).length();

        if (distance > 0.0f && distance < configuration.boid.range.separation) {
            separationSteer += separation(b, other);
            ++separationInfluenceCount;
        }

        if (distance > 0.0f && distance < configuration.boid.range.alignment) {
            alignmentSteer += alignment(b, other);
            ++alignmentInfluenceCount;
        }

        if (distance > 0.0f && distance < configuration.boid.range.cohesion) {
            cohesionSteer += cohesion(b, other);
            ++cohesionInfluenceCount;
        }
    }

    separationSteer = finalize_separation(b, separationSteer, separationInfluenceCount);
    alignmentSteer  = finalize_alignment(b, alignmentSteer, alignmentInfluenceCount);
    cohesionSteer   = finalize_cohesion(b, cohesionSteer, cohesionInfluenceCount);

    boidForces[index] = separationSteer + alignmentSteer + cohesionSteer;
}

// calculates the normal steering for the boid
void Swarm::calculateSteer(std::size_t index)
{
    Vector3f separationSteer, alignmentSteer, cohesionSteer;

    std::size_t separationInfluenceCount = 0,
                alignmentInfluenceCount  = 0,
                cohesionInfluenceCount   = 0;

    const Boid &b = boids[index];

    for (const Boid &other : boids) {
        float distance = calculateDifference(b.position, other.position).length();

        if (distance > 0.0f && distance < configuration.boid.range.separation) {
            separationSteer += separation(b, other);
            ++separationInfluenceCount;
        }

        if (distance > 0.0f && distance < configuration.boid.range.alignment) {
            alignmentSteer += alignment(b, other);
            ++alignmentInfluenceCount;
        }

        if (distance > 0.0f && distance < configuration.boid.range.cohesion) {
            cohesionSteer += cohesion(b, other);
            ++cohesionInfluenceCount;
        }
    }

    separationSteer = finalize_separation(b, separationSteer, separationInfluenceCount);
    alignmentSteer  = finalize_alignment(b, alignmentSteer, alignmentInfluenceCount);
    cohesionSteer   = finalize_cohesion(b, cohesionSteer, cohesionInfluenceCount);

    boidForces[index] = separationSteer + alignmentSteer + cohesionSteer;
}

// update the boids
// that is, calculate the new position
void Swarm::updateBoids(std::size_t first, std::size_t last)
{
    #pragma omp parallel for
    for (std::size_t index = first; index < last; ++index) {
        // update velocity
        boids[index].velocity += boidForces[index];
        boids[index].velocity  = boids[index].velocity.limitLength(configuration.boid.limit.speed);

        // update position
        boids[index].position += boids[index].velocity;

        // limit position to world border
        auto limit = [](float &value, float min, float max) {
            if (value < min)       value = max - std::abs(min - value);
            else if (value >= max) value = min + std::abs(max - value);
        };

        limit(boids[index].position.x, 0.0f, configuration.world.x);
        limit(boids[index].position.y, 0.0f, configuration.world.y);
        limit(boids[index].position.z, 0.0f, configuration.world.z);
    }
}

// calculates steering towards boids and away from other predators
void Swarm::calculatePredatorMovement(std::size_t index)
{
    Predator &p = predators[index];

    Vector3f cohesionSteer{}, // predators feel cohesion towards boids
             separationSteer{}; // and feel separation towards other predators
    
    std::size_t cohesionInfluenceCount   = 0, 
                separationInfluenceCount = 0;

    for (const Boid &other : boids) {
        // float distance = (p.position - other.position).length();
        float distance = calculateDifference(p.position, other.position).length();

        if (distance < configuration.predator.range.cohesion) {
            // then go after this boid
            cohesionSteer += cohesion(p, other);
            ++cohesionInfluenceCount;
        }
    }

    cohesionSteer = finalize_cohesion(p, cohesionSteer, cohesionInfluenceCount, configuration.predator.weight.cohesion);

    // separation towards other boids
    for (const Predator &other : predators) {
        // float distance = (p.position - other.position).length();
        float distance = calculateDifference(p.position, other.position).length();

        if (distance > 0.0f && distance < configuration.predator.range.separation) {
            separationSteer = separation(p, other);
            ++separationInfluenceCount;
        }
    }

    separationSteer = finalize_separation(p, separationSteer, separationInfluenceCount, configuration.predator.weight.separation);

    predatorForces[index] = cohesionSteer + separationSteer;
}

// update the predator
// that is, calculate the new position
void Swarm::updatePredators()
{
    #pragma omp parallel for
    for (std::size_t i = 0; i < predators.size(); ++i) {
        // update velocity
        predators[i].velocity += predatorForces[i];
        predators[i].velocity  = predators[i].velocity.limitLength(configuration.predator.limit.speed);

        // update position
        predators[i].position += predators[i].velocity;

        // limit position to world borders
        auto limit = [](float &value, const float min, const float max) {
            if (value < min)        value = max - std::abs(min - value);
            else if (value >= max)  value = min + std::abs(max - value);
        };

        limit(predators[i].position.x, 0.0f, configuration.world.x);
        limit(predators[i].position.y, 0.0f, configuration.world.y);
        limit(predators[i].position.z, 0.0f, configuration.world.z);
    }
}