// Swarm.hpp
// Defines a swarm of boids ('flock') that is broken into multiple divisions and has its
// behaviour calculated on multiple processors via MPI
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#ifndef SWARMFLOCKING_SWARM_H
#define SWARMFLOCKING_SWARM_H

#include <utility> // forward 
#include <cstdint> // size_t
#include <vector> // vector

#include "../common/Boid.h"


class Swarm
{
public:
    explicit Swarm(std::size_t boidCount, std::size_t predatorCount);
	explicit Swarm(std::size_t boidCount); // implicitly uses predatorCount = 1

    // update the whole swarm
    // this does NOT work with MPI and async communications!
    void update();

    // only update part of the swarm,
    // this does NOT work with MPI and async communications!
    void update(std::size_t first, std::size_t last);
    
    // only update part of the swarm, namely from index first to index last (excluding last)
	template <typename Fn1, typename Fn2>
    void update(std::size_t first, std::size_t last, 
                // function to call when update needs data to be synchronised 
                Fn1&& finish_sync,
                Fn2&& finish_serialisation); 

    const std::size_t boidCount;
    const std::size_t predatorCount;

    std::vector<Boid> boids;
    std::vector<Boid> predators;

private:
    bool avoidPredators(std::size_t index);
    void updateBoids(std::size_t first, std::size_t last);
    void updatePredators();
    void calculateSteer(std::size_t index);

    #pragma mark parallel optimizations
    struct partial_steer_t {
        Vector3f separation; // separation vector so far
        Vector3f alignment;  // alignment -- --
        Vector3f cohesion;   // alignment -- --
        std::size_t separationInfluenceCount;
        std::size_t alignmentInfluenceCount;
        std::size_t cohesionInfluenceCount;
        std::size_t start;   // begin of section already used
        std::size_t end;     // end of section already used (excluding end, so all indices < end are used already)
        bool evadesPredator;
    };

    void calculatePartialSteer(std::size_t index, std::size_t first, std::size_t last, partial_steer_t *out);
    void finishSteerCalculation(std::size_t index, partial_steer_t *partial);

    void calculatePredatorMovement(std::size_t index);

	std::vector<Vector3f> boidForces;
    std::vector<Vector3f> predatorForces;
};

template <typename Fn1, typename Fn2>
void Swarm::update(std::size_t first, std::size_t last,
                   Fn1&& finish_sync,
                   Fn2&& finish_serialisation)
{
    static std::vector<partial_steer_t> partials(last - first);

    // calculate partial forces
    #pragma omp parallel for
    for (std::size_t index = first; index < last; ++index) {
        if (!avoidPredators(index)) { // this is fine, because we don't synchronise the predators
            calculatePartialSteer(index, first, last, &partials[index - first]);
        } else {
            partials[index - first].evadesPredator = true;
        }
    }

    // wait for synchronisation to finish
    std::forward<Fn1>(finish_sync)();

    // finish up the calculations because now we can touch the
    // new values in the boids vector!
    #pragma omp parallel for
    for (std::size_t index = first; index < last; ++index) {
        if (!partials[index - first].evadesPredator) {
            finishSteerCalculation(index, &partials[index - first]);
        }
    }

    // calculate forces for predators
    #pragma omp parallel for
    for (std::size_t index = 0; index < predators.size(); ++index) {
        calculatePredatorMovement(index);
    }

    // now that we want to write to boids, we need to make sure all changes are persisted to disk!
    std::forward<Fn2>(finish_serialisation)();

    // ... and apply those forces
    updateBoids(first, last);
    updatePredators();
}

#endif /* SWARMFLOCKING_SWARM_H */
