// main.cpp
// Defines the applications entry-point
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include <cstdlib> // EXIT_SUCCESS, EXIT_FAILURE
#include <string> // stoi
#include <fstream> // ofstream

#include <boost/program_options.hpp>

#include <mpi.h>

#include "Simulation.h"

int main(int argc, char *argv[])
{
	MPI_Init(&argc, &argv);

    // size of swarm
    std::size_t nboids = 1024, npredators = 1;

    // steps to simulate, 0 default value for unlimited
    std::size_t steps  = 0;

    std::string output;

    namespace po = boost::program_options;
    po::options_description desc("Options");
    desc.add_options()
        ("help,h", "Print this help message")
        ("nboids", po::value<std::size_t>(&nboids), "Number of boids to simulate")
        ("npredators", po::value<std::size_t>(&npredators), "Number of predators to simulate")
        ("steps,s", po::value<std::size_t>(&steps)->required(), "Number of steps to simulate")
        ("output,o", po::value<std::string>(&output)->required(), "Specify an output file");
        // add option for no output, detailed stats after running.

    po::variables_map vm;
    try {
        po::store(po::parse_command_line(argc, argv, desc), vm);

        if (vm.count("help")) {
            std::cout << desc << std::endl; // print usage
            return EXIT_SUCCESS;
        }

        po::notify(vm);
    } catch (po::error &e) {
        std::cerr << "Erroneous usage of command line parameter: " 
                  << e.what() << std::endl << std::endl
                  << desc << std::endl;

        return EXIT_FAILURE;
    }

    MPI_File fh;

    // if output file was specified
    if (output != "") {
        const int retVal = MPI_File_open(MPI_COMM_WORLD, output.c_str(), 
                                         MPI_MODE_CREATE | MPI_MODE_WRONLY, 
                                         MPI_INFO_NULL, &fh);
        if (retVal != MPI_SUCCESS) {
            std::cerr << "Failed to open output file. Make sure you have sufficient privileges to work!" 
                      << std::endl << std::endl;
            return EXIT_FAILURE;
        }
    } else {
        std::cerr << "A valid output file is required!"
                  << std::endl << std::endl;
        return EXIT_FAILURE;
    }

    Simulation simulation{nboids,
                          npredators,
                          fh,
                          steps};
    simulation.run();

    MPI_File_close(&fh);

	MPI_Finalize();
	
    return EXIT_SUCCESS;
}
