// Visualisation.cpp
// Manages visual presentation of data using SFML. 
//
// The MIT License (MIT)
//
// Copyright (c) 2014 de.uni-hamburg.informatik.wr.papo
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all
// copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

#include "Visualisation.h"

#include <SFML/System.hpp>
#include <SFML/Window.hpp>
#include <SFML/OpenGL.hpp>

#include "../common/Configuration.h"

Visualisation::Visualisation(sf::String title, 
                             std::istream &inputSource, 
                             bool singleStepping,
                             std::size_t fps) :
    window{sf::VideoMode{800, 800, 24}, title, sf::Style::Default},
    font{},
    dataSource{inputSource},
    singleStepping{singleStepping},
    nextFrameRequested{true},
    running{true}
{
    window.setFramerateLimit(fps);

    window.setVerticalSyncEnabled(false);
    window.setActive(true);
    window.resetGLStates();

    glEnable(GL_MULTISAMPLE);
    glEnable(GL_LINE_SMOOTH);
}

void Visualisation::run()
{
    BoidBuffer::SwarmBuf buffer;
    while (window.isOpen()) {
        if (    (singleStepping  && nextFrameRequested) 
            ||  (!singleStepping && running)) {
            
            BoidBuffer::SwarmBuf oldBuf = buffer;
            buffer = dataSource.next();
            
            if (!buffer.isValid) {
                buffer = oldBuf; // continue displaying something after input is consumed
            }

            nextFrameRequested = false;
        }

        sf::Event event;
        if (window.pollEvent(event)) {
            switch (event.type) {
                case sf::Event::Closed:
                    window.close();
                    break;

                case sf::Event::Resized:
                    handleSizeEvent(event.size);
                    break;

                case sf::Event::KeyPressed:
                    handleKeyEvent(event.key);
                    break;

                case sf::Event::TextEntered:
                    handleTextEvent(event.text);
                    break;

                case sf::Event::MouseMoved:
                    handleMouseMoveEvent(event.mouseMove);
                    break;

                case sf::Event::MouseButtonPressed:
                    handleMouseButtonEvent(event.mouseButton);
                    
                case sf::Event::MouseWheelMoved:
                    handleMouseWheelEvent(event.mouseWheel);

                default:
                    break;
            }
        }

        glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        sf::Vector2u wsize = window.getSize();
        glViewport(0, 0, wsize.x, wsize.y);
        gluPerspective(60, (float)wsize.x / (float)wsize.y, 0.1f, 256.0f);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glPushMatrix();
        glTranslatef(0.0f, 0.0f, -3.5f);

        glRotatef(rotation[0], 1.f, 0.f, 0.f);
        glRotatef(rotation[1], 0.f, 1.f, 0.f);

        renderCube();
        renderSwarm(buffer);

        glPopMatrix();

        window.pushGLStates();

        window.display();
        window.popGLStates();
    }
}

void Visualisation::handleSizeEvent(sf::Event::SizeEvent sizeEvent)
{
    glViewport(0, 0, sizeEvent.width, sizeEvent.height);
}

void Visualisation::handleKeyEvent(sf::Event::KeyEvent keyEvent)
{
    switch (keyEvent.code) {
        case sf::Keyboard::Escape:
            window.close();
            break;

        case sf::Keyboard::Num1:
            rotationSpeed += 0.5f;
            break;

        case sf::Keyboard::Num2:
            if ((rotationSpeed -= 0.5f) < 0.0f) {
                rotationSpeed = 0.5f;
            }
            break;

        case sf::Keyboard::Up:
            rotation[0] -= rotationSpeed;
            break;

        case sf::Keyboard::Down:
            rotation[0] += rotationSpeed;
            break;

        case sf::Keyboard::Left:
            rotation[1] -= rotationSpeed;
            break;

        case sf::Keyboard::Right:
            rotation[1] += rotationSpeed;
            break;

        case sf::Keyboard::Space:
            if (singleStepping)
                nextFrameRequested = true;
            else 
                running = !running;
            break;

        default:
            break;
    }
}

void Visualisation::handleTextEvent(sf::Event::TextEvent textEvent)
{

}

void Visualisation::handleMouseMoveEvent(sf::Event::MouseMoveEvent mouseMoveEvent)
{

}

void Visualisation::handleMouseButtonEvent(sf::Event::MouseButtonEvent mouseButtonEvent)
{

}

void Visualisation::handleMouseWheelEvent(sf::Event::MouseWheelEvent mouseWheelEvent)
{
    
}

void Visualisation::renderSwarm(BoidBuffer::SwarmBuf &buffer)
{
    float zoom[3] = {
        configuration.world.x / 2.0f,
        configuration.world.y / 2.0f,
        configuration.world.z / 2.0f
    };

    glColor3f(0.0f, 0.0f, 1.0f);
    glPointSize(2.0f);
    
    glBegin(GL_POINTS);
        
        for (std::size_t i = 0; i < buffer.nboids; ++i) {
            Boid &current = buffer.boids[i];

            glVertex3f((current.position.x - zoom[0]) / zoom[0],
                       (current.position.y - zoom[1]) / zoom[1],
                       (current.position.z - zoom[2]) / zoom[2]);
        }

    glEnd();

    
    glColor3f(1.0f,0.0f,0.0f); // red color
    glPointSize(4.0);

    glBegin(GL_POINTS);
        
        for (std::size_t i = 0; i < buffer.npredators; ++i) {
            Predator &predator = buffer.predators[i];
            glVertex3f((predator.position.x - zoom[0]) / zoom[0],
                       (predator.position.y - zoom[1]) / zoom[1],
                       (predator.position.z - zoom[2]) / zoom[2]);
        }

    glEnd();
}

void Visualisation::renderCube()
{
    static constexpr float cubeVertices[8][3] = {
        { -1.0, -1.0,  1.0},
        { -1.0,  1.0,  1.0},
        {  1.0,  1.0,  1.0},
        {  1.0, -1.0,  1.0},
        { -1.0, -1.0, -1.0},
        { -1.0,  1.0, -1.0},
        {  1.0,  1.0, -1.0},
        {  1.0, -1.0, -1.0},
    };

    glBegin(GL_LINES);      // Draw The Cube Using lines

    glColor3f(0.0f, 0.0f, 0.0f);

    // top
    glVertex3fv(cubeVertices[0]);
    glVertex3fv(cubeVertices[1]);

    glVertex3fv(cubeVertices[1]);
    glVertex3fv(cubeVertices[2]);

    glVertex3fv(cubeVertices[2]);
    glVertex3fv(cubeVertices[3]);

    glVertex3fv(cubeVertices[3]);
    glVertex3fv(cubeVertices[0]);

    // bottom
    glVertex3fv(cubeVertices[4]);
    glVertex3fv(cubeVertices[5]);

    glVertex3fv(cubeVertices[5]);
    glVertex3fv(cubeVertices[6]);

    glVertex3fv(cubeVertices[6]);
    glVertex3fv(cubeVertices[7]);

    glVertex3fv(cubeVertices[7]);
    glVertex3fv(cubeVertices[4]);

    // connections
    glVertex3fv(cubeVertices[0]);
    glVertex3fv(cubeVertices[4]);

    glVertex3fv(cubeVertices[1]);
    glVertex3fv(cubeVertices[5]);

    glVertex3fv(cubeVertices[2]);
    glVertex3fv(cubeVertices[6]);

    glVertex3fv(cubeVertices[3]);
    glVertex3fv(cubeVertices[7]);
    
    glEnd(); // End Drawing The Cube
}
